# coding=UTF-8

import wx, wx.html, subprocess
from nts.ntsData import *
from nts.ntsDocs import *
import wx.lib.dialogs
import  wx.stc as stc

# for the edge normal color
NORMAL = '#E5E5E5'
# for the edge modified color
ACTIVE = '#33FF00'
MODIFIED = '#FF6103'
# for search highlighting
SEARCHF = '#000000'
SEARCHB = '#FFE303'

wx.SetDefaultPyEncoding(encoding)

from pkg_resources import resource_filename

options = get_opts()
filters = get_filters(options)

ID_BUTTON=100
ID_SPLITTER=300

try:
    import markdown
    import markdown.extensions.abbr
    import markdown.extensions.footnotes
    import markdown.extensions.tables
    import markdown.extensions.def_list
    import markdown.extensions.fenced_code
    import markdown.extensions.headerid
    markdown_version = markdown.version
    has_markdown = True
except:
    has_markdown = False
    markdown_version = 'none'
try:
    from docutils import __version__ as docutils_version
    from docutils.core import publish_string
    has_docutils = True
except:
    docutils_version = 'none'
    has_docutils = False

if pandoc:
    # get pandoc version
    try:
        p = subprocess.Popen("%s -v | head -n1" % pandoc, shell=True, stdout = subprocess.PIPE)
        ret = p.stdout.read()
        pandoc_version = ret.split()[1]
        has_pandoc = True
    except:
        pandoc = None
        pandoc_version = "none"
        has_pandoc = False
else:
    pandoc_version = "none"
    has_pandoc = False

def sysinfo():
    from platform import python_version as pv
    wxv = "%s.%s.%s" % (wx.MAJOR_VERSION, 
                wx.MINOR_VERSION, wx.RELEASE_VERSION)
    sysinfo = "platform %s; python %s; wx(Python) %s" % (sys.platform, pv(), wxv)
    return(sysinfo)

def newer_alert():
    ok, res = get_newer()
    if ok and res:
        msg = "%s\n\n  Set 'newer = False' in %s\n  to prevent this check for newer releases." % (res, ntsrc)
        dlg = wx.MessageDialog(None, msg,  'nts', wx.OK )
        dlg.ShowModal() 
    return(True)


if "wxMac" in wx.PlatformInfo:
    notepad_type = 'png'
    notepad = "NTS_128.png"
elif "wxMSW" in wx.PlatformInfo:
    notepad_type = 'ico'
    notepad = "NTS.ico"
elif "wxGTK" in wx.PlatformInfo:
    notepad_type = 'png'
    notepad = "NTS_32.png"
else:
    notepad_type = ''
    notepad = ''

if notepad:
    try:
        nts_notepad = resource_filename(__name__, notepad)
    except:
        nts_notepad = notepad

class MyHtmlWindow(wx.html.HtmlWindow):
    def __init__(self, parent, id, pos = wx.DefaultPosition, size=wx.DefaultSize, style=0):
        wx.html.HtmlWindow.__init__(self, parent, id, pos, size,
            style=wx.BORDER_SUNKEN)
        if "gtk2" in wx.PlatformInfo:
            self.SetStandardFonts(htmlfont+4, '', '')
        else:
            self.SetFonts('', '', [i for i in range(htmlfont,
                htmlfont+13, 2)])

    def AcceptsFocus(self, *args, **kwargs):
        return False

class NTShtml(wx.Dialog):
    def __init__(self, parent = None, size=wx.DefaultSize, page = ''):
        wx.Dialog.__init__(
            self, parent, -1, 'nts', size=size, pos=wx.DefaultPosition, 
            style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)
        self.parent = parent
        sizer = wx.BoxSizer(wx.VERTICAL)
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.html = MyHtmlWindow(self, -1, size=size)
        self.html.Bind(wx.EVT_CHAR, self.OnChar)
        self.html.SetBorders(0)
        self.SetMinSize(size)
        self.page = page
        self.html.SetPage(self.page)
        self.printer = wx.html.HtmlEasyPrinting()
        self.printer.SetFonts('', '', [i for i in range(htmlprintfont,
            htmlprintfont+13, 2)])
        self.printdata = self.printer.GetPrintData()
        self.printdata.SetColour(False)
        self.Bind(wx.EVT_CHAR, self.OnChar)
        vbox.Add(self.html, 1, wx.GROW | wx.ALL, 4)
        #  sizer.Add(vbox, 0, wx.GROW|wx.ALIGN_CENTER_VERTICAL|wx.ALL, 1)
        sizer.Add(vbox, 1, wx.GROW|wx.ALL, 1)
        btnsizer = wx.BoxSizer(wx.HORIZONTAL)
        self.btn2 = wx.Button(self, wx.ID_PRINT)
        self.btn2.SetLabel("Print")
        self.btn2.Bind(wx.EVT_BUTTON, self.OnPrint)
        btnsizer.Add(self.btn2, 0, wx.RIGHT, 3)
        self.btn1 = wx.Button(self, wx.ID_CANCEL)
        self.btn1.Bind(wx.EVT_BUTTON, self.OnCancel)
        self.btn1.SetDefault()
        btnsizer.Add(self.btn1, 0, wx.LEFT, 3)
        sizer.Add(btnsizer, 0, wx.ALIGN_CENTER | wx.BOTTOM, 5)
        self.SetSizer(sizer)
        sizer.Fit(self)

    def OnEnter(self, event):
        event = wx.CommandEvent(wx.wxEVT_COMMAND_BUTTON_CLICKED, self.btn1.GetId()) 
        self.btn1.Command(event) 

    def OnPrint(self, event):
        self.printer.SetHeader(
          '<center><font size="+1">%s</font></center>' %
          self.html.GetOpenedPageTitle())
        self.printer.SetFooter(
                '<center>Page @PAGENUM@ of @PAGESCNT@</center>')
        try:
            self.printer.PrintText(self.page)
        except:
            pass

    def OnChar(self, event):
        keycode = event.GetKeyCode()
        if keycode in [27, 17, ord('q'), ord('Q')]:  # Escape, Ctrl-Q, q or Q quit
            self.OnCancel(event)
        elif keycode in [16, ord('p'), ord('P')]: # Ctrl-P, p or P print
            self.OnPrint(event)
        else:
            event.Skip()

    def OnCancel(self, event):
        event.Skip()
        self.Destroy()

class MyTaskBarIcon(wx.TaskBarIcon):
    TBMENU_SHOW  = wx.NewId()
    TBMENU_HIDE  = wx.NewId()
    TBMENU_CLOSE = wx.NewId()

    def __init__(self, frame):
        wx.TaskBarIcon.__init__(self)
        self.frame = frame
        if notepad_type == 'ico':
            self.SetIcon(wx.Icon(nts_notepad, wx.BITMAP_TYPE_ICO), 'nts')
        elif notepad_type == 'png':
            self.SetIcon(wx.Icon(nts_notepad, wx.BITMAP_TYPE_PNG), 'nts')

        self.Bind(wx.EVT_TASKBAR_LEFT_DCLICK, self.OnTaskBarActivate)
        self.Bind(wx.EVT_MENU, self.OnTaskBarActivate, id=self.TBMENU_SHOW)
        self.Bind(wx.EVT_MENU, self.OnTaskBarDeActivate, id=self.TBMENU_HIDE)
        self.Bind(wx.EVT_MENU, self.OnTaskBarClose, id=self.TBMENU_CLOSE)

    def CreatePopupMenu(self):
        menu = wx.Menu()
        menu.Append(self.TBMENU_SHOW, "Show")
        menu.Append(self.TBMENU_HIDE, "Hide")
        menu.AppendSeparator()
        menu.Append(self.TBMENU_CLOSE, "Close")
        return menu

    def OnTaskBarClose(self, event):
        self.frame.Close()

    def OnTaskBarActivate(self, event):
        if "wxMSW" in wx.PlatformInfo and self.frame.IsIconized():
            self.frame.Iconize(False)
        if not self.frame.IsShown():
            self.frame.Show(True)
        self.frame.Raise()

    def OnTaskBarDeActivate(self, event):
        if self.frame.IsShown():
            self.frame.Hide()

# vars = {'version': version, 'copyright' : copyright, 'ntsinfo' : ntsinfo(), 'sysinfo' : sysinfo(), 'markup' : markup(), 'ntstxt' : ntstxt, 'ntsenc' : ntsenc, 'ntsexport' : os.path.join(ntsexport, 'export.rst'), 'ntsrc' : ntsrc, 'ntsdata' : ntsdata, 'grandchild' :  os.path.join(ntsdata,'parent', 'child', 'grandchild.txt')}

class MyTree(wx.TreeCtrl):
    def __init__(self, parent, id, position, size, style):
        wx.TreeCtrl.__init__(self, parent, id, position, size, style)
        # lofl item: (description, content, tags, id, path) 
        self.displayed = []
        self.rel_path = ''
        self.line_beg = 0
        self.show_id = None
        self.MakeTree(options, filters)

    def SetActive(self, rel_path, line_beg):
        self.rel_path = rel_path
        self.line_beg = line_beg

    def MakeTree(self, options, filters):
        self.DeleteAllItems()
        root = self.AddRoot('root')
        self.root = root
        self.id2item = {}
        lofl = make_tree(options, filters)
        paths = []
        parentof = {}
        self.id2note = {}
        self.displayed = []
        self.show_id = None
        for note in lofl:
            id = note[3]
            self.id2note[note[3]] = note
            self.displayed.append((note[0], note[3]))
            path = note[4]
            # the last component of path is the leaf (.txt)
            for i in range(len(path)+1):
                part = path[:i]
                if part and part not in paths:
                    full_node = ""
                    for node in part:
                        full_node = os.path.join(full_node, node)
                    paths.append(part)
                    description = note[0]
                    if i == 1: 
                        parentof[part[-1]] = self.AppendItem(root,
                            part[-1])
                    else: 
                        parentof[part[-1]] = self.AppendItem(
                                parentof[part[-2]], part[-1])
                    if i == len(path):
                        self.SetPyData(parentof[part[-1]], 
                            (note[0], '', '', (note[3][0],0,0), full_node))
                    else:
                        self.SetPyData(parentof[part[-1]], 
                            (note[0], '', '', (), full_node))

            description = "%s" % note[0]
            parentof[description] = self.AppendItem(
                    parentof[part[-1]], description)
            self.id2item[note[3]] = parentof[description]
            if (self.rel_path and id[0] == self.rel_path and 
                    id[1] == self.line_beg):
                # save this item to make visible in OnReload
                self.show_id = parentof[description]
            text = "%s" % ("\n".join(note[1]))

            if note[2] and 'none' not in note[2]:
                tag_str = "(%s)" % ', '.join(note[2])
            else:
                tag_str = ""
            content = (description, text, tag_str, id, path)
            self.SetPyData(parentof[description], content)
        num = len(self.id2item.keys())
        if len(lofl) == 1:
            self.num_notes = "%s note" % num
        else:
            self.num_notes = "%s notes" % num
        self.SetPyData(root, ('', '', '', (), ""))


class MySTC(stc.StyledTextCtrl):
    def __init__(self, parent, ID):
        stc.StyledTextCtrl.__init__(self, parent, ID, size=(80,40))

        self.Bind(stc.EVT_STC_MODIFIED, self.OnModified)

        self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)
        self.editable = False
        self.modified = False
        self.StyleSetSpec(stc.STC_STYLE_LINENUMBER, "back:%s" % NORMAL)
        # style 1 is for the search highlight
        self.StyleSetSpec(1, "fore:%s,back:%s" % (SEARCHF, SEARCHB))
        self.SetCaretWidth(2)
        self.SetMarginWidth(0,0)
        self.SetMarginWidth(1,4)
        self.SetMarginType(1, stc.STC_MARGIN_SYMBOL)
        self.SetMarginMask(1, 0)  
        self.SetMarginType(2, stc.STC_MASK_FOLDERS)
        self.SetMarginWidth(2, 0)
        self.SetMargins(4,1)

        self.MarkerDefine(stc.STC_MARKNUM_FOLDER, stc.STC_MARK_BOXPLUS,  "white", "black")


    def OnDestroy(self, evt):
        wx.TheClipboard.Flush()
        evt.Skip()

    def ChangeValue(self, txt):
        self.SetReadOnly(False)
        self.SetText(txt)
        self.SetReadOnly(True)

    def GetValue(self):
        return self.GetText()

    def SetEditable(self, bool):
        self.editable = bool
        if bool:
            self.SetReadOnly(False)
            self.StyleSetSpec(stc.STC_STYLE_LINENUMBER, "back:%s" % ACTIVE)
        else:
            self.SetReadOnly(True)
            self.StyleSetSpec(stc.STC_STYLE_LINENUMBER, "back:%s" % NORMAL)

    def IsEditable(self):
        return self.editable

    def IsModified(self):
        return self.CanUndo()

    def SetStyle(self, b, e, ignore):
        self.StartStyling(b, 0xff)
        self.SetStyling(e-b, 1)

    def OnModified(self, evt):
        if self.IsEditable():
            if self.CanUndo():
                self.StyleSetSpec(stc.STC_STYLE_LINENUMBER, "back:%s" % MODIFIED)
            else:
                self.StyleSetSpec(stc.STC_STYLE_LINENUMBER, "back:%s" % ACTIVE)
        else:
            self.StyleSetSpec(stc.STC_STYLE_LINENUMBER, "back:%s" % NORMAL)

    def getModType(self, evt):
        modType = evt.GetModificationType()
        st = ""
        table = [(stc.STC_MOD_INSERTTEXT, "InsertText"),
                 (stc.STC_MOD_DELETETEXT, "DeleteText"),
                 (stc.STC_MOD_CHANGESTYLE, "ChangeStyle"),
                 (stc.STC_MOD_CHANGEFOLD, "ChangeFold"),
                 (stc.STC_PERFORMED_USER, "UserFlag"),
                 (stc.STC_PERFORMED_UNDO, "Undo"),
                 (stc.STC_PERFORMED_REDO, "Redo"),
                 (stc.STC_LASTSTEPINUNDOREDO, "Last-Undo/Redo"),
                 (stc.STC_MOD_CHANGEMARKER, "ChangeMarker"),
                 (stc.STC_MOD_BEFOREINSERT, "B4-Insert"),
                 (stc.STC_MOD_BEFOREDELETE, "B4-Delete")
                 ]
        for flag,text in table:
            if flag & modType:
                st = st + text + " "
        if not st:
            st = 'UNKNOWN'
        return st

    def AcceptsFocus(self, *args, **kwargs):
        print "accepts focus?", self.editable
        return self.editable


class MyFrame(wx.Frame):
    def __init__(self, parent, id, title):
        wx.Frame.__init__(self, parent, id, title,
                          wx.DefaultPosition, wx.Size(840, 400))
        self.SetBackgroundColour('#EEEEEE')
        if notepad_type == 'ico':
            self.SetIcon(wx.Icon(nts_notepad, wx.BITMAP_TYPE_ICO))
            self.tbicon = MyTaskBarIcon(self)
        elif notepad_type == 'png':
            self.SetIcon(wx.Icon(nts_notepad, wx.BITMAP_TYPE_PNG))
            self.tbicon = MyTaskBarIcon(self)
        self.Bind(wx.EVT_ICONIZE, self.OnIconify)
        self.Bind(wx.EVT_CLOSE, self.OnQuit)
        # outline tree
        if mono_tree:
            tfont = wx.Font(font_size, wx.MODERN, wx.NORMAL, wx.NORMAL)
        else:
            tfont = wx.Font(font_size, wx.DEFAULT, wx.NORMAL, wx.NORMAL)
        # display panel view mode
        if mono_display:
            dfont = wx.Font(font_size, wx.MODERN, wx.NORMAL, wx.NORMAL)
        else:
            dfont = wx.Font(font_size, wx.DEFAULT, wx.NORMAL, wx.NORMAL)
        # display panel edit mode
        if mono_edit:
            efont = wx.Font(font_size, wx.MODERN, wx.NORMAL, wx.NORMAL)
        else:
            efont = wx.Font(font_size, wx.DEFAULT, wx.NORMAL, wx.NORMAL)
        self.dfont = dfont
        self.efont = efont
        self.markup = None
        # labels in the option panel
        lfont = wx.Font(font_size - 1, wx.DEFAULT, wx.NORMAL, wx.NORMAL)
        # buttons in the option panel
        bfont = wx.Font(font_size - 1, wx.DEFAULT, wx.NORMAL, wx.NORMAL)
        # entry fields in the option panel
        ofont = wx.Font(font_size, wx.DEFAULT, wx.NORMAL, wx.NORMAL)
        # monospaced font for the help display
        mfont = wx.Font(font_size, wx.MODERN, wx.NORMAL, wx.NORMAL)
        self.mfont = mfont
        self.splitter = wx.SplitterWindow(self, ID_SPLITTER,
                style = wx.SP_LIVE_UPDATE)
        leftPanel = wx.Panel(self.splitter, -1, 
                style = wx.LC_LIST | wx.NO_FULL_REPAINT_ON_RESIZE)
        leftBox = wx.BoxSizer(wx.VERTICAL)
        self.hofh = {}
        self.content = ''
        self.local_filters = get_filters({})
        self.hofh = load_data({})
        self.tree = MyTree(leftPanel, -1, wx.DefaultPosition, 
            wx.DefaultSize,
            wx.TR_HIDE_ROOT
            | wx.TR_HAS_BUTTONS
            )
        self.tree.SetBackgroundColour("white")
        self.tree.Bind(wx.EVT_CHAR, self.OnChar)
        leftBox.Add(self.tree, 1, wx.EXPAND | wx.LEFT | wx.TOP | wx.BOTTOM,
                4)
        self.tree.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.OnSelActivated)
        self.tree.Bind(wx.EVT_TREE_SEL_CHANGED, self.OnSelChanged)
        self.tree.Bind(wx.EVT_TREE_SEL_CHANGING, self.OnSelChanging)
        self.tree.SetFont(tfont)
        leftPanel.SetSizer(leftBox)
        rightPanel = wx.Panel(self.splitter, -1, size = wx.DefaultSize,
            style=wx.SP_3D)
        rightBox = wx.BoxSizer(wx.VERTICAL)
        self.display = MySTC(rightPanel, -1)
        self.display.SetWrapMode(True)
        self.display.Bind(wx.EVT_CHAR, self.OnDisplayChar)
        self.display.Bind(wx.EVT_KEY_DOWN, self.OnDisplayChar)
        self.display.Bind(wx.EVT_LEFT_DCLICK, self.OnSelActivated)
        self.display.StyleSetFont(stc.STC_STYLE_DEFAULT, dfont)
        rightBox.Add(self.display, 1, wx.EXPAND | wx.ALL, 4)
        rightPanel.SetSizer(rightBox)
        self.splitter.SplitVertically(leftPanel, rightPanel, 270)

        self.optionbar = wx.BoxSizer(wx.HORIZONTAL)
        button1 = wx.StaticText(self, ID_BUTTON + 1, "^O ")
        button1.SetFont(lfont)
        self.gb = wx.Choice(self, ID_BUTTON + 2, choices = ['p', 't'])
        self.gb.SetFont(bfont)
        button3 = wx.StaticText(self, ID_BUTTON + 3, "^P ")
        button3.SetFont(lfont)
        self.pf = wx.TextCtrl(self, ID_BUTTON + 4, "")
        self.pf.SetFont(ofont)
        self.pf.Bind(wx.EVT_CHAR, self.OnOptChar)
        button5 = wx.StaticText(self, ID_BUTTON + 5, "^T ")
        button5.SetFont(lfont)
        self.tf = wx.TextCtrl(self, ID_BUTTON + 6, "")
        self.tf.SetFont(ofont)
        self.tf.Bind(wx.EVT_CHAR, self.OnOptChar)
        button7 = wx.StaticText(self, ID_BUTTON + 7, "^F ")
        button7.SetFont(lfont)
        self.ff = wx.TextCtrl(self, ID_BUTTON +8, "")
        self.ff.SetFont(ofont)
        self.ff.Bind(wx.EVT_CHAR, self.OnOptChar)
        # apply
        self.eb = wx.Button(self, ID_BUTTON +9, "F4: APPLY", style = wx.BU_EXACTFIT)
        self.eb.SetFont(bfont)
        self.Bind(wx.EVT_BUTTON, self.OnEnter, self.eb)
        self.cb = wx.Button(self, ID_BUTTON +10, "F3: CLEAR", style = wx.BU_EXACTFIT)
        self.cb.SetFont(bfont)
        self.Bind(wx.EVT_BUTTON, self.OnClear, self.cb)
        self.hb = wx.Button(self, ID_BUTTON +11, "F1: HELP", style = wx.BU_EXACTFIT)
        self.hb.SetFont(bfont)
        self.Bind(wx.EVT_BUTTON, self.OnHelp, self.hb)
        s = 8
        self.optionbar.Add(button1, 0, wx.EXPAND | wx.LEFT, s)
        self.optionbar.Add(self.gb, 0, wx.EXPAND | wx.RIGHT, s)
        self.optionbar.Add(button3, 0, wx.EXPAND | wx.LEFT, s)
        self.optionbar.Add(self.pf, 1, wx.EXPAND | wx.RIGHT, s)
        self.optionbar.Add(button5, 0, wx.EXPAND | wx.LEFT, s)
        self.optionbar.Add(self.tf, 1, wx.EXPAND | wx.RIGHT, s)
        self.optionbar.Add(button7, 0, wx.EXPAND | wx.LEFT, s)
        self.optionbar.Add(self.ff, 1, wx.EXPAND | wx.RIGHT, s)
        self.optionbar.Add(self.hb, 0, wx.EXPAND | wx.RIGHT, s)
        self.optionbar.Add(self.cb, 0, wx.EXPAND | wx.RIGHT, s)
        self.optionbar.Add(self.eb, 0, wx.EXPAND  | wx.RIGHT, s)

        self.vbox = wx.BoxSizer(wx.VERTICAL)
        self.vbox.Add(self.splitter, 1, wx.EXPAND)
        self.vbox.Add(self.optionbar, 0, wx.EXPAND | wx.BOTTOM, 2)
        self.SetSizer(self.vbox)

        self.tree.SetFocus()
        self.treeFocused = True
        self.modified = False
        self.leaf_selected = False
        self.node_selected = False
        self.selected_id = ()
        self.active_id = ()
        nts_count = self.CountChildren(self.tree.root)
        self.SetTitle("nts (%s)" % self.tree.num_notes)
        self.content = nts_count
        self.display.ChangeValue(nts_count)
        self.Centre()
        if newer:
            newer_alert()
        self.tree.SetFocus()

    def CountChildren(self, item):
        children = self.tree.GetChildrenCount(item, recursively=False)
        if item == self.tree.root:
            head = "Displayed:"
        else:
            head = "The selected item has:"
        if children == 1:
             nts_children= "%d child" %1
        else:
            nts_children = "%d children" % children
        other = self.tree.GetChildrenCount(item)
        descendants = other - children
        if descendants == 1:
             nts_descendants= " and %d other descendant" % 1
        else:
            nts_descendants = " and %d other descendants" % descendants
        if descendants > 0:
            nts_count = "%s%s" % (nts_children, nts_descendants)
        else:
            nts_count = "%s" % (nts_children)
        return(nts_count)

    def OnIconify(self, event):
        if "wxMSW" in wx.PlatformInfo:
            self.Hide()

    def OnSelChanged(self, event):
        item =  event.GetItem()
        self.selected = item
        self.leaf_selected = ()
        self.node_selected = ()
        content = self.tree.GetPyData(item)
        #  content: 0:description, 1:text, 2:tag_str, 3:id, 4:path)
        if content[3]:
            self.selected_id = content[3]
            self.leaf_selected = content[3]
            if content[2]:
                tag_str = " %s" % content[2]
            else:
                tag_str = ""
            if content[3][1]:
                detail_str = "%s\n\n-- %s%s" % (content[1], 
                        ':'.join(map(str, content[3])),
                        tag_str)

                self.display.ChangeValue("%s" % detail_str)
                if (self.local_filters['find_regex'] and 
                    not self.local_filters['neg_find']):
                    for m in self.local_filters['find_regex'].finditer(detail_str):
                        b, e = m.span()
                        self.display.SetStyle(b, e, wx.TextAttr(wx.RED))
            else:
                nts_count = self.CountChildren(item)
                self.display.ChangeValue(nts_count)
        else:
            self.selected_id = (content[4], 0, 0)
            self.node_selected = content[4]
            nts_count = self.CountChildren(item)
            self.display.ChangeValue(nts_count)
        self.content = self.display.GetValue()

    def OnSelActivated(self, event):
        item = self.selected
        if self.display.IsEditable():
            event.Skip()
            return()
        # Display the selected item text in the text widget
        item = self.tree.GetPyData(item)
        if item[1]:
            if item[2]:
                tag_str = " %s" % item[2]
            else:
                tag_str = ""
            self.display.StyleSetFont(stc.STC_STYLE_DEFAULT, self.efont)
            self.display.ChangeValue("+ %s%s\n%s" % (item[0], tag_str, item[1]))
            self.active_id = item[3]
            self.selected_id = item[3]
            self.treeFocused = False
            self.display.SetFocus()
            self.display.DocumentStart()
            self.display.EmptyUndoBuffer()
            self.display.SetEditable(True)

    def DeActivate(self):
        if self.display.IsEditable():
            if self.display.IsModified():
                dlg = wx.MessageDialog(None, 'abandon changes?',  'nts',
                    wx.YES_NO | wx.YES_DEFAULT )
                if dlg.ShowModal() == wx.ID_NO:
                    return () 
            self.display.SetEditable(False)
        self.display.StyleSetFont(stc.STC_STYLE_DEFAULT, self.dfont)
        self.display.ChangeValue(self.content)
        self.treeFocused = True
        self.active_id = ()
        self.display.SetEditable(False)
        self.tree.SetFocus()

    def OnSelChanging(self, event):
        item =  event.GetItem()
        if self.display.IsEditable() and self.display.IsModified():
            event.Veto()
            dlg = wx.MessageDialog(None, 'abandon changes?',  'nts',
                    wx.YES_NO | wx.YES_DEFAULT )
            if dlg.ShowModal() == wx.ID_NO:
                self.display.SetFocus()
                self.treeFocused = False
                return ()
        self.display.SetEditable(False)
        self.treeFocused = True
        self.active_id = ()
        self.tree.SetFocus()
        self.display.ChangeValue(self.content)
        event.Skip()

    def ToggleFocus(self):
        if self.treeFocused:
            self.display.SetFocus()
            self.treeFocused = False
        else:
            self.DeActivate()

    def OnChar(self, event):
        keycode = event.GetKeyCode()        # for arrow keys
        ukeycode = event.GetUnicodeKey()    # for control keys
        shift = event.ShiftDown()
        control = event.ControlDown()
        modifier = event.GetModifiers()
        #  print "OnChar", modifier, keycode, ukeycode
        if keycode == 27:               # ESC
            self.DeActivate()
        elif keycode == wx.WXK_LEFT:
            if (control and self.selected and 
                    self.tree.IsSelected(self.selected)):
                parent = self.tree.GetItemParent(self.selected)
                if parent == self.tree.root:
                    self.tree.CollapseAllChildren(self.selected)
                else:
                    self.tree.CollapseAllChildren(parent)
            elif (self.selected and self.tree.IsSelected(self.selected) and
                self.tree.ItemHasChildren(self.selected)):
                self.tree.CollapseAllChildren(self.selected)
            event.Skip()
        elif control and keycode == wx.WXK_RIGHT:
            if (self.tree.IsSelected(self.selected) and
                self.tree.ItemHasChildren(self.selected)):
                self.tree.ExpandAllChildren(self.selected)
            event.Skip()
        elif modifier == 2 and ukeycode == 3:           # Ctrl-C convert
            self.OnConvert()

        elif modifier == 2 and ukeycode == 8:           # Ctrl-H html
            ids = self.getIDs()
            if ids:
                self.OnHTML(ids)
        elif ukeycode == 12:                             # ^L log file
            self.OnInfo()
        elif control and ukeycode == 13 and self.leaf_selected:  # Ctrl-M move
            # Return generates 13 as well but not control
			self.OnMove()
        elif ukeycode == 14:           # Ctrl-N add new note
            self.OnNew()
        elif ukeycode == 24:           # Ctrl-X export
            ids = self.getIDs()
            if ids:
                str = self.OnExport(ids)
        elif ukeycode == 17:           # Ctrl-Q quit
            self.OnQuit()
        elif ukeycode == 18:           # Ctrl-R reload
            self.hofh = {}
            self.ReloadData()
        elif ukeycode == 21:           # Ctrl-U unselect all
            self.UnselectAll()
        elif modifier == 0 and ukeycode in [8, 127] and self.leaf_selected:
            self.OnDelete()
        elif keycode == wx.WXK_TAB:
            self.ToggleFocus()
        elif keycode == wx.WXK_F1: 
            self.OnHelp(event)
        elif keycode == wx.WXK_F2: 
            self.OnAbout(event)
        elif keycode == wx.WXK_F3: 
            self.OnClear(event)
        elif keycode == wx.WXK_F4: 
            self.OnEnter(event)
        elif keycode == wx.WXK_F5: 
            self.OnOpen()
        elif ukeycode in [15, 16, 20, 6]: # ^O, ^P, ^T, ^F
            #  print "calling setfocus", ukeycode
            self.setfocus(event, ukeycode)
        else:
            event.Skip()

    def OnDisplayChar(self, event):
        keycode = event.GetKeyCode()
        ukeycode = event.GetUnicodeKey()    # for control keys
        shift = event.ShiftDown()
        control = event.ControlDown()
        modifier = event.GetModifiers()
        #  print "OnDisplayChar", modifier, control, shift, keycode, ukeycode
        if keycode == wx.WXK_F1: 
            self.OnHelp(event)
        elif keycode == wx.WXK_F2: 
            self.OnAbout(event)
        elif keycode == 27 and not self.display.IsModified():
            self.DeActivate()
        else:
            if self.display.IsEditable():
                if ukeycode == 17:           # Ctrl-Q quit
                    self.DeActivate()
                elif control and ukeycode == 19:           # Ctrl-S save
                    self.OnSave(event, False)
                else:
                    event.Skip()
            else:
                if keycode == wx.WXK_TAB:
                    self.DeActivate()
                elif keycode in [13, wx.WXK_DELETE, wx.WXK_BACK]: # 13 = return
                    self.OnSelActivated(event)
                else:
                    self.DeActivate()
                    self.OnChar(event)

    def OnOptChar(self, event):
        keycode = event.GetKeyCode()
        shift = event.ShiftDown()
        control = event.ControlDown()
        if keycode in [15, 16, 20, 6]: # ^O, ^P, ^T, ^S 71, 80, 84, 83
            self.setfocus(event, keycode+64)
        elif keycode == 13:    # enter
            if control:
                self.OnClear(event)
            else:
                self.OnEnter(event)
        elif keycode == 17:           # Ctrl-Q quit
            self.OnQuit()
        if keycode == 27:               # ESC
            if self.treeFocused:
                self.tree.SetFocus()
            else:
                self.display.SetFocus()
        elif keycode == wx.WXK_F1: 
            self.OnHelp(event)
        elif keycode == wx.WXK_F2: 
            self.OnAbout(event)
        elif keycode == wx.WXK_F3: 
            self.OnClear(event)
        elif keycode == wx.WXK_F4: 
            self.OnEnter(event)
        else:
            event.Skip()

    def OnAbout(self, event):
        info = wx.AboutDialogInfo()
        info.Name = "nts"
        info.Version = "%s" % version
        info.Copyright = "(C) %s Daniel A. Graham" % copyright
        info.Description = description
        info.WebSite = ("http://www.duke.edu/~dgraham/NTS", 
                "nts home page")
        info.Developers = [ "Daniel A. Graham <daniel.graham@duke.edu>", ]

        info.License = license
        wx.AboutBox(info)

    def OnEnter(self, evt):
        local_options = options
        local_options['outlineby']  = ['p', 't'][self.gb.GetCurrentSelection()]
        local_options['path']  = self.pf.GetValue()
        tags = self.tf.GetValue()
        if tags:
            m = parens_regex.match(tags)
            if m:
                parts = m.group(1).split(',')
                parts = [x.strip() for x in parts]
            else:
                parts = [tags]
            local_options['tag']  = parts 
        else:
            local_options['tag'] = []
        local_options['find']  = self.ff.GetValue()
        local_filters = get_filters(local_options)
        self.local_filters = local_filters
        self.tree.MakeTree(local_options, local_filters)
        nts_count = self.CountChildren(self.tree.root)
        self.SetTitle("nts (%s)" % self.tree.num_notes)
        self.display.ChangeValue(nts_count)
        if local_options['path'] or local_options['tag'] or local_options['find']:
            self.tree.ExpandAll()
        self.tree.SetFocus()

    def ReloadData(self):
        local_options = options
        self.hofh = load_data(self.hofh)
        self.tree.MakeTree(options, filters)
        nts_count = self.CountChildren(self.tree.root)
        self.SetTitle("nts (%s)" % self.tree.num_notes)
        self.display.ChangeValue(nts_count)
        self.tree.SetFocus()
        if self.tree.show_id:
            self.tree.EnsureVisible(self.tree.show_id)
            self.tree.SelectItem(self.tree.show_id)


    def UnselectAll(self):
        nts_count = self.CountChildren(self.tree.root)
        self.display.ChangeValue(nts_count)
        self.treeFocused = True
        self.active_id = ()
        self.selected_id = ()
        self.leaf_selected = False
        self.node_selected = False
        self.tree.SetFocus()
        self.tree.UnselectAll()

    def OnClear(self, evt):
        gb = self.gb.SetSelection(0)
        self.pf.SetValue('')
        self.tf.SetValue('')
        self.ff.SetValue('')

    def setfocus(self, event, key):
        if key == 15:
            # toggle outlineby
            old_gb = self.gb.GetCurrentSelection()
            new_gb = (old_gb + 1) % 2
            self.gb.SetSelection(new_gb)
            self.OnEnter(event)
        elif key == 16:
            self.pf.SetFocus()
        elif key == 20:
            self.tf.SetFocus()
        elif key == 6:
            self.ff.SetFocus()
        else:
            pass

    def OnSave(self, event, must_exist = False):
        if not self.display.IsEditable() or not self.display.IsModified():
            return()
        #  note_id: (relative_path, beginning_line, ending_line) 
        id = self.active_id
        lines = self.ContentLines()
        if id:
            # this is an existing note
            rel_path = id[0]
            line_beg = noteReplace(id, lines)
        else:
            # this is a new note
            file = self.getFile("Append new note to", must_exist)
            if file:
                root, ext = os.path.splitext(file)
                if ext not in [ntstxt, ntsenc]: 
                    # make ntstxt the default
                    file = "%s%s" % (root, ntstxt)
                    if ext:
                        # only confirm use of default for bad extensions
                        dlg = wx.MessageDialog(None, 
                            'The file extension "%s" was provided but should either\nbe "%s" or "%s". The default extension "%s" and file\n\n            "%s"\n\nwill be used.  Is this OK?' % (ext, ntstxt, ntsenc, ntstxt, file),  
                            'nts',
                            wx.YES_NO | wx.YES_DEFAULT )
                        if dlg.ShowModal() == wx.ID_NO:
                                return(False) 
                line_beg = noteAdd(file, lines)
                rel_path = relpath(file, ntsdata)
            else:
                line_beg = False
        if line_beg:
            #  self.display.SetModified(False)
            self.display.SetEditable(False)
            #  self.display.SetBackgroundColour('white')
            self.treeFocused = True
            self.active_id = ()
            self.tree.SetActive(rel_path, line_beg)
            self.ReloadData()
            self.tree.SetFocus()
        return(line_beg)

    def getSelection(self, prompt, choices, dflt=None):
        dlg = wx.SingleChoiceDialog(self, prompt, 'nts',
                choices, wx.CHOICEDLG_STYLE)
        if dflt:
            dlg.SetSelection(dflt)
        if dlg.ShowModal() == wx.ID_OK:
            return dlg.GetStringSelection()
        else:
            return ""

    def getFile(self, prompt, must_exist = False):
        file = None
        wildcard = "nts plain text files (*%s)|*%s|nts encoded files (*%s)|*%s|all files (*.*)|*.*"  % (ntstxt, ntstxt, ntsenc, ntsenc) 
        if must_exist:
            style = wx.FD_OPEN | wx.FD_FILE_MUST_EXIST
        else:
            #  style = wx.FD_SAVE
            style = wx.FD_SAVE
            #  style = wx.FD_OPEN
        filedlg = wx.FileDialog(
                self, message = prompt,
                defaultDir = ntsdata,
                defaultFile="%s" % quick,
                wildcard = wildcard,
                style=style
        )
        #  filedlg.SetWildcard(wildcard)
        if filedlg.ShowModal() == wx.ID_OK:
            file = filedlg.GetPath()
            if file:
                return(os.path.join(ntsdata, file))
        return()

    def getIDs(self):
        choices = ['ALL OF THE FOLLOWING']
        ids = []
        selected = []
        for note in self.tree.displayed:
            choices.append("%s [%s]" % (note[0], ":".join(map(str,note[1]))))
            ids.append(note[1])
        dlg = wx.MultiChoiceDialog(self, "notes to include", 'nts',
               choices)
        if self.selected_id:
            if self.selected_id in ids:
                # add one because inserting ALL makes the choices index 1
                # greater than the ids index
                selected.append(ids.index(self.selected_id)+1)
            else:
                # a note is not selected so get the path to the node and 
                # append all descendent notes
                for i in range(len(choices)):
                    if self.selected_id[0] in choices[i]:
                        selected.append(i)
            dlg.SetSelections(selected)
        if dlg.ShowModal() == wx.ID_OK:
            indices = dlg.GetSelections()
        else:
            indices = []
        if 0 in indices:
            return(ids)
        else:
            return([ids[i-1] for i in indices])

    def MarkUp(self, ids):
        out = []
        if ids:
            if markup in ['md', 'pd']:
                # out = ['# nts #']
                out = []
            else:
                out = ['nts', '===', '']
            for id in ids:
                note = self.tree.id2note[id]
                if markup in ['md', 'pd']:
                    out.append("### %s ###" % note[0])
                    # out.append('')
                else: 
                    out.append(note[0])
                    out.append('-'*len(note[0]))
                    out.append('')
                for line in note[1]:
                    out.append("%s" % line)
                out.append('')
        return(out)

    def OnExport(self, ids):
        '''Export selected notes in md or rst format'''
        out = self.MarkUp(ids)
        if out:
            export_file = os.path.join(ntsexport, 'export.%s' % markup)
            fo = codecs.open(export_file, 'w', encoding, 'replace')
            fo.writelines(["%s\n" % x for x in out])
            fo.close()

    def OnInfo(self):
        f = os.path.join(ntsexport, 'messages.log')
        fo = open
        fo = codecs.open(f, 'r', encoding, 'replace')
        messages = fo.readlines()
        fo.close()
        message = "".join(messages)
        dlg = wx.lib.dialogs.ScrolledMessageDialog(self, message,
                'nts message log')
        dlg.ShowModal()

    def CheckMarkup(self):
        if not self.markup:
            md = "markdown %s" % markdown_version
            du = "docutils %s" % docutils_version
            pd = "pandoc %s" % pandoc_version
            self.markup = "%s; %s; %s" % (du, md, pd)

    def OnHTML(self, ids):
        '''Show selected notes in HTML window'''
        html = ''
        out = self.MarkUp(ids)
        if out:
            txt = "\n".join(out)
        else:
            return()
        if markup == 'pd' and has_pandoc: 
            p = subprocess.Popen([pandoc, '-t html'], shell = True, 
                    stdout = subprocess.PIPE, stdin = subprocess.PIPE, 
                    stderr = subprocess.PIPE)
            html, err = p.communicate(txt.encode(encoding))
        elif markup == 'md' and has_markdown: 
            html = markdown.markdown(txt, 
                    ['abbr', 'footnotes', 'tables', 
                        'def_list', 'fenced_code', 'headerid'])
        elif markup == 'rst' and has_docutils:
            if stylesheet:
                customcss = {'embed_stylesheet':False, 
                'stylesheet_path':stylesheet}
                html = publish_string(source=txt, writer_name='html',
                    settings_overrides=customcss)
            else:
                html = publish_string(source=txt, writer_name='html')
        dlg = NTShtml(self, size=(400,300), page = html)
        dlg.Show(True)

    def OnNew(self):
        self.display.ChangeValue("+ ")
        self.display.SetEditable(True)
        self.active_id = ()
        self.treeFocused = False
        self.display.SetFocus()
        self.display.GotoPos(2)

    def OnDelete(self):
        dlg = wx.MessageDialog(None, 'delete note?',  'nts',
            wx.YES_NO | wx.YES_DEFAULT )
        if dlg.ShowModal() == wx.ID_NO:
                return ()
        res = noteDelete(self.leaf_selected)
        self.tree.SetActive('', 0)
        if res == 2:
            del self.hofh[self.leaf_selected[0]]
        if res:
            self.active_id = ()
            self.tree.SetFocus()
            self.ReloadData()
        return(True)

    def OnConvert(self):
        orig_file = self.getFile("File to convert", True)
        if orig_file:
            pathname, ext = os.path.splitext(orig_file)
            if ext == ntsenc:
                new_ext = ntstxt
                tocrypted = False
            else:
                new_ext = ntsenc
                tocrypted = True
            new_file = "%s%s" % (pathname, new_ext)
            dlg = wx.MessageDialog(None, 'convert %s\nto %s?' % (orig_file, new_file),  'nts',
                wx.YES_NO | wx.YES_DEFAULT )
            if dlg.ShowModal() == wx.ID_NO:
                return () 
            res = ConvertFile(orig_file, new_file, tocrypted, True)
            if res:
                rel_path = relpath(orig_file, ntsdata)
                del self.hofh[rel_path]
                self.active_id = ()
                self.tree.SetFocus()
                self.ReloadData()


    def OnMove(self):
        id = self.leaf_selected
        file = self.getFile("Move note to", False)
        if file:
            item = self.tree.id2note[id]
            if item[2]:
                tag_str = " (%s)" % ", ".join(item[2])
            else:
                tag_str = ""
            lines = ["\n", "+ %s%s\n" % (item[0], tag_str)]
            for line in item[1]:
                lines.append("%s\n" % line)
            added = noteAdd(file, lines)
            deleted = 0
            if added:
                res = noteDelete(self.leaf_selected)
                if res == 2:
                    del self.hofh[self.leaf_selected[0]]
                if res:
                    self.active_id = ()
                    self.tree.SetFocus()
            if res or added:
                self.ReloadData()

    def OnOpen(self):
        file = None
        temp = None
        if self.node_selected: 
            dir = self.node_selected
            dir = full_path(dir)
            try:
                OpenWithDefault(dir)
            except:
                print "Error: could not open '%s'" % dir
        elif self.leaf_selected:
            file, beg, end = self.leaf_selected
            file = full_path(file)
        if file:
            backup(file)
            pathname, ext = os.path.splitext(file)
            if ext == ntsenc:  # encoded, need a plain text copy
                temp = "%s%s" % (pathname, ntstxt)
                ConvertFile(file, temp, False, True) # removes orig
                command = editcmd % {'e': editor, 'n': beg, 'f': temp}
                os.system(command)
                ConvertFile(temp, file, True, True) # removes temp
            else: # plain text
                command = editcmd % {'e': editor, 'n': beg, 'f': file}
                os.system(command)
            self.ReloadData()

    def ContentLines(self):
        # get rid of the trailing new lines
        content = self.display.GetValue()
        lines = content.split('\n')
        while not lines[-1].rstrip('\n'):
            lines.pop()
        # and put one back
        lines.append('\n')
        return(["%s" % x.rstrip('\n') for x in lines])

    def OnHelp(self, event):
        self.CheckMarkup()
        vars = {'version': version, 'copyright' : copyright, 'ntsinfo' : ntsinfo(), 'sysinfo' : sysinfo(), 'markup' : self.markup, 'ntstxt' : ntstxt, 'ntsenc' : ntsenc, 'ntsexport' : os.path.join(ntsexport, 'export.rst'), 'ntsrc' : ntsrc, 'ntsdata' : ntsdata, 'grandchild' :  os.path.join(ntsdata,'parent', 'child', 'grandchild.txt')}
        dlg = NTShtml(self, size=(600,340), page = help % vars )
        dlg.Show(True)

    def OnQuit(self, evnt=None):
        self.tbicon.Destroy()
        self.Destroy()

class MyApp(wx.App):

    def OnInit(self):
        frame = MyFrame(None, -1, 'nts')
        frame.Show(True)
        self.SetTopWindow(frame)
        return True

def main():
    app = MyApp(0)
    app.MainLoop()

if __name__ == '__main__':
    main()


description = """\
nts (note taking simplified) provides a simple format for using 
text files to store notes, a command line interface for viewing
notes in a variety of convenient ways and a wx(python)-based
GUI for creating and modifying notes as well as viewing them.
"""

# license = "GNU General Public License (GPL)"
license = """\
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version. 
[ http://www.gnu.org/licenses/gpl.html ]

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty
of MERCHANTabILITY or FITNESS FOR A PARTICULAR PURPOSE. See
the GNU General Public License for more details.
"""


help = """\
        <title>nts: note taking simplified</title>
<center>
nts version %(version)s. Copyright %(copyright)s Daniel A Graham. All rights reserved.
<br>
%(ntsinfo)s
<br>
%(sysinfo)s
<br>
%(markup)s
</center>
<pre>
Non editing shortcuts:
    F1                show this help information.
    F2                show 'about' information for nts.
    F3                clear option panel settings.
    F4                apply option panel settings.
    F5                if a note or leaf (notes file) is selected, open the
                      file using the external editor. Otherwise, open the node
                      (directory) using the system default application,
                      usually the file manager.
    Ctrl C            convert file. Select a file from a dialog and, if plain
                      text, then base-64 encode it and otherwise, if encoded,
                      then decode it. In either case, make the appropriate
                      change in the file extension.
    Ctrl F            move focus to the option panel f (FIND) field.
    Ctrl H            display notes as HTML with an option to print. Select
                      notes to be displayed from a list with defaults
                      corresponding to the current selection. Requires the
                      python package docutils. See 'Markup' below for more
                      information.
    Ctrl L            show nts message log with details of any problems
                      experienced in the last loading of the data files.
    Ctrl M            move the selected note it to a file to be selected 
                      from a file dialog. This dialog also allows creating
                      new files and folders.
    Ctrl N            create a new note. When finished use Ctrl-S, Shift
                      Ctrl-S or Alt-S to save it.
    Ctrl O            cycle the option panel o (OUTLINEBY) setting between
                      p)aths and t)ags.
    Ctrl P            move focus to the option panel p (PATH) field.
    Ctrl Q            quit nts.
    Ctrl R            reload data files.
    Ctrl T            move focus to the option panel t (TAG) field.
    Ctrl U            unselect all items.
    Ctrl X            export notes. Select notes to be exported from a list
                      with defaults corresponding to the current selection.
                      Selected notes will be exported as a restructured text
                      file to
                          %(ntsexport)s
                      (specified in %(ntsrc)s).

    Left Arrow        collapse any children and then move to the parent.
    Ctrl Left Arrow   move to parent and then collapse all children.
    Right Arrow       expand any children and then move down in the outline.
    Ctrl Right Arrow  expand entire branch and then move down in the outline.
    Up/Down Arrows    move up or down in outline without expanding or
                      collapsing any items.

    Double click      with a note selected, open the note for editing.
    Return            in the option panel PATH, TAG or FIND fields, apply
                      option panel settings. In the outline panel with a note
                      selected, open the note for editing.
    Tab               cycle focus between the outline and display panels.
    Escape            cancel editing a note. Warn if the note has been
                      modified.
    Delete            delete the selected note.

Editing shortcuts:
    Ctrl A                  Select all the text in the document
    Ctrl C                  Copy the selection to the clipboard
    Ctrl L                  Cut the line containing the cursor.
    Ctrl+Shift L            Delete the line containing the cursor.
    Ctrl Q                  Quit editing note. Prompt for confirmation if 
                            note is modified.
    Ctrl S                  Save changes to a note. If editing an existing 
                            note, replace it with the modified version. If 
                            creating a new note, append it to a file to be
                            chosen from a file dialog. This dialog also
                            allows creating new files and folders. 
    Ctrl T                  Switch the current line with the previous.
    Ctrl U                  Transform the selection to lower case.
    Ctrl+Shift U            Transform the selection to upper case.
    Ctrl V                  Paste the contents of the clipboard into the
                            document replacing the selection.
    Ctrl X                  Cut the selection to the clipboard
    Ctrl Y                  Redoes the next action on the undo history
    Ctrl Z                  Undo one action in the undo history
    Down Arrow              Move cursor down one line.
    Shift Down Arrow        Move cursor down one line extending selection 
                            to new cursor position.
    Ctrl Down Arrow         Scroll the document down, keeping the cursor
                            visible.
    Up Arrow                Move cursor up one line.
    Shift Up Arrow          Move cursor up one line extending selection to 
                            new cursor position.
    Ctrl Up Arrow           Scroll the document up, keeping the cursor 
                            visible.
    Left Arrow              Move cursor left one character.
    Shift Left Arrow        Move cursor left one character extending 
                            selection to new cursor position.
    Ctrl Left Arrow         Move cursor left one word.
    Ctrl+Shift Left Arrow   Move cursor left one word extending selection 
                            to new cursor position.
    Alt Left Arrow          Move to the previous change in capitalisation.
    Alt+Shift Left Arrow    Move to the previous change in capitalisation.
                            extending selection to new cursor position.
    Right Arrow             Move cursor right one character.
    Shift Right Arrow       Move cursor right one character extending 
                            selection to new cursor position.
    Ctrl Right Arrow        Move cursor right one word.
    Ctrl+Shift Right Arrow  Move cursor right one word extending selection 
                            to new cursor position.
    Alt Right Arrow         Move to the next change in capitalisation.
    Alt+Shift Right Arrow   Move to the next change in capitalisation
                            extending selection to new cursor position.
    Home                    Move cursor to before first visible character 
                            on line. If already there move to first character 
                            on line.
    Shift Home              Like Home but extending selection to new 
                            cursor position.
    Ctrl Home               Move cursor to first position in document.
    Ctrl+Shift Home         Move cursor to first position in document 
                            extending selection to new cursor position.
    Alt Home                Move cursor to first position on display line.
    Alt+Shift Home          Move cursor to first position on display line
                            extending selection to new cursor position.
    End                     Move cursor to last position on line.
    Shift End               Move cursor to last position on line extending
                            selection to new cursor position.
    Ctrl End                Move cursor to last position in document.
    Ctrl+Shift End          Move cursor to last position in document 
                            extending selection to new cursor position.
    Alt End                 Move cursor to last position on display line.
    Alt+Shift End           Move cursor to last position on display line
                            extending selection to new cursor position.
    Page up                 Move cursor one page up.
    Shift Page up           Move cursor one page up extending selection to 
                            new cursor position.
    Page down               Move cursor one page down.
    Shift Page down         Move cursor one page down extending selection 
                            to new cursor position.
    Shift Backspace         Cut the selection to the clipboard.
    Ctrl Delete             Delete the word to the left of the cursor.
    Ctrl Backspace          Delete the word to the right of the cursor
    Ctrl+Shift Delete       Delete backwards from the current position to 
                            the beginning of the line.
    Ctrl+Shift Backspace    Delete forwards from the current position to 
                            the end of the line.
    Insert                  Switch from insert to overtype mode or the 
                            reverse.
    Shift Insert            Paste the contents of the clipboard into the 
                            document replacing the selection.
    Ctrl Insert             Copy the selection to the clipboard
    Tab                     If selection is empty or all on one line 
                            replace the selection with a tab character. If
                            more than one line is selected, indent the 
                            lines.
    Shift Tab               Dedent the selected lines
    Return                  Insert a new line, may use a CRLF, CR or LF
                            depending on EOL mode.
    Shift Return            Insert a new line, may use a CRLF, CR or LF
                            depending on EOL mode.

Option panel settings:
    o OUTLINEBY  An element from [p, t] where:
                p: outline by path
                t: outline by tag
                Default: p.
    p PATH       Regular expression. Include items with paths matching PATH
                (ignoring case). Prepend an exclamation mark, i.e., use !PATH
                rather than PATH, to include items which do NOT have contexts
                matching PATH.
    t TAG        Regular expression. Include items with tags matching TAG
                (ignoring case). Prepend an exclamation mark, i.e., use !TAG
                rather than TAG, to include items which do NOT have tags
                matching TAG.
    f FIND       Regular expression. Include items containing FIND (ignoring
                case) in the note description or note text. Prepend an
                exclamation mark, i.e., use !FIND rather than FIND, to include
                notes which do NOT have descriptions or note texts matching
                FIND.

Making changes:
    To edit an existing note, either double-click on the note or select the 
    note and then press Return. Make your changes to the note in the display 
    panel and then press Ctrl-S when you are finished to replace the original 
    note with your modified version.

    To create a new note, press Ctrl-N and then create your note in the 
    display panel. When you are finished, press Ctrl-S to save the 
    note either to an existing notes file or to a new file. The dialog will 
    offer the opportunity to create new folders if necessary.

    To delete a note, select the note and then press Delete. You will be
    prompted to confirm the deletion. If the last note in a file were removed,
    then the file itself would be deleted.

    To move a note, select the note and then press Ctrl-M. You will be 
    prompted to select a notes file. The note will then be appended to the new 
    file and deleted from the existing file. If the last note in a file were 
    moved, then the file itself would be deleted. 

    When making any of the changes listed above, nts will first make a backup
    of the existing file - see 'Rotating backup files' below. 

    You can also make changes to notes files using the editor you specified in
    your ~/.nts/rc file. Press F5 with a note selected and the file will be
    opened in your editor with the cursor on the relevant line. When the file
    containing the selected note is encoded, a temporary, plain text version
    of the file is created for editing. When finished editing the original
    file is replaced with an encoded version of the saved temporary file and
    the temporary file is then removed.

    Please use your favorite file manager to reorganize your nts data
    directory structure. If you press F5 on a node with any children, your
    system default application will be called with the relevant path. Normally
    this will open your file manager at the relevent directory.

Hierarchical notes:
    Notes files either have the extension '%(ntstxt)s' (plain text) or the
    extension '%(ntsenc)s' (base 64 encoded) and are located in or below 

        %(ntsdata)s

    Note that these settings reflect those in the rc file currently in use,
    '%(ntsrc)s'.

    Both file types support unicode characters with normal, readable display
    both in the GUI and in command line output. The base 64 encoding is
    intended to provide only VERY LIGHT WEIGHT protection (obfuscation) for
    encoded files outside nts.

    The directory structure below ntsdata provides the hierarchy for your
    notes. E.g., suppose you have the notes file:

            %(grandchild)s

    with the following content:

        ----------- begin grandchild.txt ----------------------
        + note a (tag 1, tag 2)
        the body of my first note

        + note b (tag 2, tag 3)
        the body of my second note
        ----------- end grandchild.txt ------------------------

    Then when outlining by **path** you would see:

        parent
            child
                grandchild
                    note a
                    note b

    and when outlining by **tag** you would see:

        tag 1
            note a
        tag 2
            note a
            note b
        tag 3
            note b

    Each notes file can contain one or more notes using the following format
    for each::

        + note title (optional tags)
        one or more lines containing the body of the note
        with all white space preserved.

    In the note title, the '+' must be in the first column. If given, tags
    must be comma separated and enclosed in parentheses. White space in the
    note body is preserved but whitespace between notes is ignored. The note
    body continues until another note line or the end of the file is reached.

Date and time stamps:
    If you enter ".d" (without the quotes) anywhere in note that you are 
    creating or editing in the GUI, it will be expanded to the current date in 
    YYYY-MM-DD format when you save the note. Similarly, ".dt" will be expanded 
    to the current date and time.

    You can associate dates (or datetimes) with your notes by entering ".d" 
    (or ".dt") as one of the tags for the note. You can then see your "dated" 
    notes grouped and sorted by date (or datetime) by using "-o t -t 
    \d{4}-\d{2}-\d{2}" in either the CLI or the GUI.

Rotating backup files:
    A backup is made of any file before nts makes any changes to it. For
    example, before saving a change to the base 64 encoded file,
    'mynotes.enc', the exising file would first be copied to '.mynotes.bk1'.
    If '.mynotes.bk1' already exists and it is more than one day old, it would
    first be moved to '.mynotes.bk2'. Similarly, if '.mynotes.bk2' already
    exists, then it would be first be moved to '.mynotes.bk3' and so forth. In
    this way, up to 'numbaks' (3 by default) rotating backups of are kept with
    '.bk1' the most recent.

    The process is similar for a plain text file but the copy is encoded
    before saving. Thus all backups are base 64 encoded.

Markup:
    Either 'markdown' or 'restructured text' markup can be used in the body of
    notes. Moreover, by setting either

        markup = "md"  

    or 

        markup = "rst"

    in your nts rc file, nts will provide consistent markup when exporting
    selected notes. Further, if markdown (or docutils for restructured text)
    is installed on your system, you will be able to display selected notes as
    html with an option to print.

    There are many similarities between the two types of markup, e.g., *this
    would be emphasized* and **this would be bold-faced** under either. More
    generally, markdown is somewhat simpler to use but also somewhat less
    powerful. For details, see

          http://daringfireball.net/projects/markdown/syntax

          http://docutils.sourceforge.net/docs/user/rst/quickref.html

    Using the standard tools that come with python's docutils module, rst
    output can be easily converted to a number of formats including HTML,
    Latex and OpenOffice ODF. Similarly, md output can be converted to other
    formats using 'pandoc':

          http://johnmacfarlane.net/pandoc/ 
</pre>
"""



import sys, datetime, os, os.path, fnmatch, shutil, re, subprocess
import locale, codecs
# import base64               # for crypt
from base64 import b64encode, b64decode
from dateutil.tz import tzlocal, tzutc

from nts.ntsRC import *
from nts.ntsVersion import version
from platform import system
platform = system()


try:
    # try using the user's default settings.
    locale.setlocale(locale.LC_ALL, '')
except:
    # use the current setting for locale.LC_ALL
    locale.setlocale(locale.LC_ALL, None)

if editor and not EDITOR:
    EDITOR = editor
if editcmd and not EDITCMD:
    EDITCMD = editcmd

note_regex = re.compile(r'^\+\s+([^\(]+)\s*(\(([^\)]*)\))?\s*$')
comment_regex = re.compile(r'^\s*#')
empty_regex = re.compile(r'^\s*$')
leadingspaces = re.compile(r'^(\s*)\S.*')

regex_end = re.compile(r'^\s*!\s*$')
regex_date = re.compile(r'\.d\b')
regex_datetime = re.compile(r'\.dt\b')
parens_regex = re.compile(r'^\s*\((.*)\)\s*$')

messages =[]
warning = "" 
options = {}


today = datetime.date.today()
lastyear = str(int(today.strftime("%Y"))-1)
thismonth = today.strftime("%m")
thisyear = today.strftime("%Y")

if thisyear == '2010':
    copyright = '2010'
else:
    copyright = '2010-%s' % thisyear

def ntsinfo():
    ntsinfo = "ntsdir: %s; ntsdata: %s" % (ntsdir, ntsdata)
    return(ntsinfo)

def get_newer():
    global version
    import socket
    timeout = 10
    socket.setdefaulttimeout(timeout)
    # strip the '-x' from experimental versions
    version = (version.split('-'))[0]
    from urllib2 import Request, urlopen, URLError
    req = Request("http://www.duke.edu/~dgraham/NTS/version.txt")
    try:
        response = urlopen(req)
    except URLError, e:
        if hasattr(e, 'reason'):
            msg = """\
We failed to reach a server in checking for a newer version of nts.
Reason: %s.""" % e.reason
        elif hasattr(e, 'code'):
            msg = """\
The server couldn\'t fulfill the request.
Error code: %s.""" % e.code
        return(1, msg)
    else:
        # everything is fine

        vstr = response.read().strip()
        if int(version) < int(vstr): 
            return(1, 'A newer release of nts, version %s, is available.' % (vstr))
        else:
            return(0, 'You have the latest version of nts.')

try:
    from os.path import relpath
except ImportError: # python < 2.6
    from os.path import curdir, abspath, sep, commonprefix, pardir, join
    def relpath(path, start=curdir):
        """Return a relative version of a path"""
        if not path:
            raise ValueError("no path specified")
        start_list = abspath(start).split(sep)
        path_list = abspath(path).split(sep)
        # Work out how much of the filepath is shared by start and path.
        i = len(commonprefix([start_list, path_list]))
        rel_list = [pardir] * (len(start_list)-i) + path_list[i:]
        if not rel_list:
            return curdir
        return join(*rel_list)

def OpenWithDefault(path):
    if platform in ('Windows', 'Microsoft'):
        os.startfile(path)
    elif platform == 'Darwin':
        subprocess.Popen('/usr/bin/open' + " %s" % path, shell = True)
    else:
        subprocess.Popen('xdg-open' + " %s" % path, shell = True)

def get_filelist(topdir=ntsdata, txtpattern='[!.]*%s' % ntstxt,
    encpattern='[!.]*%s' % ntsenc):
    """yield the list of files in topdir and its subdirectories whose names match pattern."""
    filelist = []
    for path, subdirs, names in os.walk(topdir):
        for name in names:
            if (fnmatch.fnmatch(name, txtpattern) or
                fnmatch.fnmatch(name,encpattern)):
                full_path = os.path.join(path,name)
                rel_path = relpath(full_path, topdir)
                filelist.append((rel_path, full_path))
    return(filelist)

def crypt(str):
    str = str.rstrip('\n')
    # print "crypt orig: '%s'" % str
    if str.endswith("="):
        # shouldn't happen
        return(str)
    if str:
        # add a starting ! to reserve leading white space
        str = "!%s" % str
    retval = b64encode(str.encode(encoding)).rstrip('\n')
    #  retval = b64encode(str).rstrip('\n')
    if retval and not retval.endswith('='):
        # make sure we have an ending =
        retval = "%s=" % retval
    return(retval)

def decrypt(str):
    str = str.rstrip('\n')
    try:
        retval = unicode(b64decode(str), encoding)
    except:
        print "bad string: '%s'" % str
        retval = None
    if retval:
        # drop the starting !
        retval = retval[1:]
    return(retval)

def CryptLines(lines):
    return(["%s\n" % crypt(x) for x in lines])

def DeCryptLines(lines):
    return(["%s\n" % decrypt(x) for x in lines])

def ConvertFile(orig_file, new_file, tocrypted = True, remove_orig = True):
    '''If plain text convert to encrypted, else convert encrypted to plain text'''
    fo = codecs.open(orig_file, 'r', encoding, 'replace')
    f = fo.readlines()
    fo.close()
    lines = []
    for line in f:
        line = line.rstrip('\n')
        if tocrypted:
            lines.append(crypt(line))
        else:
            lines.append(decrypt(line))
    fo = codecs.open(new_file, 'w', encoding, 'replace')
    fo.writelines(["%s\n" % x for x in lines])
    fo.close()
    if remove_orig:
        os.remove(orig_file)
    return(True)

def process_file(rel_path, full_path):
    # print "process_file", rel_path
    global messages
    i2n = {}
    i2p = {}
    r2p = {}
    r2i = {}
    t2i = {}
    pathname, ext = os.path.splitext(rel_path)
    if ext == ntsenc:
        encrypt = True
    else:
        encrypt = False
    context = []
    head = pathname
    while head:
        head, tail = os.path.split(head)
        context.insert(0, tail)
    path = tuple(context)
    r2p[rel_path] = path
    fo = codecs.open(full_path, 'r', encoding, 'replace')
    f = fo.readlines()
    fo.close()
    if not f:
        messages.append("skipping %s:\n    could not obtain content" 
                % rel_path)
        return()
    line = f[0].strip()
    if encrypt:
        line = decrypt(line)
    if not note_regex.match(line):
        messages.append("%s:\n    the first line does not begin a note" 
                % rel_path)
        return()
    note = {}
    line_num = 0
    try:
        for line in f:
            line = line.rstrip('\n')
            if encrypt:
                line = decrypt(line)
            line_num += 1
            if note_regex.match(line):
                if note:
                    # end previous note
                    note_end = line_num - 1
                    note_id = (rel_path, note_beg, note_end) 
                    # print "set id", note_id
                    note['id'] = note_id
                    i2n[note_id] = note
                    i2p[note_id] = path
                    for tag in note['tags']:
                        t2i.setdefault(tag, []).append(note_id)
                    r2i.setdefault(rel_path, []).append(note_id)
                    while len(note['text']) > 0 and len(note['text'][-1]) == 0:
                        note['text'].pop()
                    note = {}
                m = note_regex.match(line)
                note = {'path' : path}
                note['description'] = m.group(1).strip()
                note['text'] = []
                note_beg = line_num
                if m.group(3):
                    try:
                        tags = [x.strip() for x in m.group(3).split(',')]
                        note['tags'] = tags
                    except:
                        note['tags'] = ['none']
                else:
                    note['tags'] = ['none']
            else: # must be a note text line
                if note:
                    note['text'].append(line)
    except:
        note = {}
        messages.append("exception raised processing %s\n    %s" 
                % (rel_path, sys.exc_info()))
        return()
    if note:
        note_end = line_num
        note_id = (rel_path, note_beg, note_end) 
        note['id'] = note_id
        while len(note['text']) > 0 and len(note['text'][-1]) == 0:
            note['text'].pop()
        i2n[note_id] = note
        i2p[note_id] = path
        r2i.setdefault(rel_path, []).append(note_id)
        if note['tags']:
            for tag in note['tags']:
                t2i.setdefault(tag, []).append(note_id)
        else:
            t2i.setdefault('none', []).append(note_id)
    return(i2n, i2p, r2p, r2i, t2i)

def get_relpath(exp = ''):
    if exp:
        regex = re.compile(r'%s' % exp.decode(encoding), re.IGNORECASE)
    else:
        regex = ''
    n = 0
    hsh = {}
    for rel_path, full_path in get_filelist():
        if regex and not regex.match(rel_path):
            continue
        n += 1
        hsh[n] = rel_path
        print "[%s] %s" % (n, rel_path)
    print """\
----------------------------------------------------------------
    Enter the number of an existing file or the relative path
    of an new file to which the new note should be appended.
----------------------------------------------------------------\
    """
    rp = raw_input('> ')
    if len(rp) > 0:
        try:
            rp = hsh[int(rp)]
        except:
            pass
    else:
        rp = ''
    return(rp)

def process_data(hofh = {}):
    # note = (description, note_lines, tags, id) 
    # where id = (rel_path, beg_line, end_line)
    global messages
    messages = []
    id2note = {}
    id2pth = {}
    rel2pth = {}
    rel2ids = {}
    tag2ids = {}
    for rel_path, full_path in get_filelist():
        pathname, ext = os.path.splitext(rel_path)
        mtime = os.path.getmtime(full_path)
        if rel_path not in hofh or hofh[rel_path][0] < mtime:
            lst = process_file(rel_path, full_path)
            if not lst:
                # the problem should already have been appended to messages
                continue
            hofh[rel_path] = [mtime, lst]
    for mtime, [i2n, i2p, r2p, r2i, t2i] in hofh.values():
        id2note.update(i2n)
        id2pth.update(i2p)
        rel2pth.update(r2p)
        for key in r2i:
            rel2ids.setdefault(key, []).extend(r2i[key])
        for key in t2i:
            tag2ids.setdefault(key, []).extend(t2i[key])
    if messages:
        for message in messages:
            print("%s" % message)
    return(id2note, id2pth, rel2pth, rel2ids, tag2ids, hofh)

def make_tree(options, filters):
    # lofl: [nodes] 
    # where each node = [title, [content], [tags], (id), (path)]
    display = ["%s: %s" % (key, options[key]) for key in options.keys() if options[key] ]
    if display:
        display.sort()
        hdr = "; ".join(display) 
    lofl = []
    if options['outlineby'] == 't':
        tags = tag2ids.keys()
        if tags:
            tags.sort()
            for tag in tags:
                if filters['tag_regex']:
                    match = False
                    for nt, tr in filters['tag_regex']:
                        r = tr.search(tag)
                        t_res = (r and r.group(0).strip())
                        if (nt and t_res):
                            continue
                        if not nt and not t_res:
                            continue
                        match = True
                    if not match:
                        continue
                notes = []
                for id in tag2ids[tag]:
                    note = id2note[id]
                    if filter_note(filters, note):
                        path = [tag]
                        # put a lowercase description first for sorting
                        notes.append([note['description'].lower(),
                            note['description'], note['text'], note['tags'],
                            note['id'], path])
                # sort by the first, lower case term
                # notes.sort(key = lambda a:a[0])
                notes.sort()
                # now remove the first, lower case term
                lofl.extend([note[1:] for note in notes])
    else:
        rels = rel2pth.keys()
        if rels:
            # hopefully unicode will not be a problem in the paths
            rels.sort(key = str.lower)
            for rel in rels:
                if filters['path_regex']:
                    r = filters['path_regex'].search(rel)
                    c_res = (r and r.group(0).strip())
                    if (filters['neg_path'] and c_res):
                        continue
                    if not filters['neg_path'] and not c_res:
                        continue
                matching_ids = []
                notes = []
                for id in rel2ids[rel]:
                    note = id2note[id]
                    if filter_note(filters, note):
                        rel_path, beg_line, end_line = note['id'] # id
                        path = split_path(rel_path)
                        # put a lowercase description first for sorting
                        notes.append([note['description'].lower(),
                            note['description'], note['text'], note['tags'],
                            note['id'], path])
                # sort by the first, lower case term
                # notes.sort(key = lambda a:a[0])
                notes.sort()
                # now remove the first, lower case term
                lofl.extend([note[1:] for note in notes])
    return(lofl)


def split_path(rel_path):
    path = []
    head, ext = os.path.splitext(rel_path)
    while head:
        head, tail = os.path.split(head)
        path.insert(0, tail)
    return(path)

def print_tree(tree):
    display = int(options['display'])
    number = int(options['number'])
    tags = options['outlineby'] == 't'
    if 'lines' in options:
        num_lines = options['lines']
    else:
        num_lines = 0
    if options['edit']:
        edit = options['edit']
    else:
        edit = 0
    if options['editfile']:
        editfile = options['editfile']
    else:
        editfile = 0
    if  options['remove']:
        remove = options['remove']
        #  show everything for confirmation
        display = 7
        num_lines = 4
    else:
        remove = 0
    arglist = []
    if options['args']:
        for arg in options['args']:
            if ':' in arg:
                b, e = map(int, arg.split(':'))
                for i in range(b, e+1):
                    arglist.append(i)
            else:
                arglist.append(int(arg))
    # note = [title, [content], [tags], (id), (path)]
    # where id = (rel_path, beg_line, end_line)
    paths = []
    tab = '  '
    last_tag = ''
    indent = 0
    num_indent = 0
    body_indent = 0
    title_indent = 0
    info_indent = 0
    if display in [3,7]:
        body_indent = 1
    if display in [7]:
        info_indent = 1
    if display == 2 and number == 0:
        ret = True
    else:
        ret = False
    num = 0
    if tags:
        if number:
            num_indent = 1
        else:
            title_indent = 1
        body_indent += 1
        info_indent += 1
    for note in tree:
        num += 1
        if arglist and num not in arglist:
            continue
        if edit and num != edit:
            continue
        if editfile and num != editfile:
            continue
        if remove and num != remove:
            continue
        rel_path, beg_line, end_line = note[3] # id
        path = note[4]
        if display in [1, 3, 5, 7]:
            title = "%s%s" % (tab*title_indent, note[0])
        else:
            title = ""
        if display in [2,3,6,7]:
            if note[1]:
                #  get leading white space
                first = note[1][0]
                ws = len(first) - len(first.lstrip())
                if options['lines']:
                    num_lines = int(options['lines'])
                    omitted = len(note[1])-num_lines
                    if omitted > 1:
                        lines = note[1][:num_lines]
                        while len(lines) > 0 and len(lines[-1].strip()) == 0:
                            lines.pop()
                        lines.append('... (%s omitted lines)' % omitted)
                    else:
                        lines = note[1]

                else:
                    lines = note[1]
                if ws:
                    ws_regex = re.compile(r'^\s{1,%d}' % ws)
                    body = ["%s%s" % (tab*body_indent, ws_regex.sub('', x)) 
                        for x in lines]
                else:
                    body = ["%s%s" % (tab*body_indent, x) 
                        for x in lines]
            else:
                print('error: missing body for note %s' % note[0])
        else:
            body = []
        if display in [4,5,6,7]:
            if note[2] and note[2][0] != 'none':
                tag_str = " (%s)" % ", ".join(note[2])
            else:
                tag_str = ""
            info = "%s-- %s%s" % (tab*info_indent,
                    ':'.join(map(str,note[3])), tag_str)
        else:
            info = ""
        if options['edit']:
            edit_note(note[3])
            return()
        if options['editfile']:
            edit_file(note[3])
            return()
        if remove:
            print"""\
---------------------------------------------------------------------
            The following note is to be removed.
---------------------------------------------------------------------"""
        if tags:
            if path[0] != last_tag:
                last_tag = path[0]
                print "%s" % path[0]
        if number:
            print "%s[%s]" % (tab*num_indent, num),
        if title:
            print "%s%s" % (tab*title_indent, title)
        if body:
            for line in body:
                print line
        if info:
            print info
        if remove:
            print"""\
---------------------------------------------------------------------"""
            confirm = raw_input('Delete this note? [yN] ')
            if confirm.upper() == 'Y':
                noteDelete(note[3])
            else:
                print "Cancelled"
            return()
        if ret:
            print ""


def has(hash, key):
    "Return true if key and hash and key in hash and hash[key]"
    try:
        # return(key and hash and key in hash and hash[key] != None)
        return(key and hash and key in hash and 
            hash[key] not in [None, '', {}])
    except:
        print 'except', key, hash
        return(False)

def full_path(rel_path):
    return os.path.join(ntsdata, rel_path)

def get_filters(options):
    filter = {}
    if has(options,'find'):
        if options['find'][0] == '!':
            filter['neg_find'] = True
            if type(options['find']) == str:
                filter['find_regex'] = re.compile(r'%s' %
                    options['find'][1:].decode(encoding), re.IGNORECASE)
            else:
                filter['find_regex'] = re.compile(r'%s' %
                    options['find'][1:], re.IGNORECASE)
        else:
            filter['neg_find'] = False
            if type(options['find']) == str:
                filter['find_regex'] = re.compile(r'%s' %
                    options['find'].decode(encoding), re.IGNORECASE)
            else:
                filter['find_regex'] = re.compile(r'%s' %
                    options['find'], re.IGNORECASE)
    else:
        filter['neg_find'] = False
        filter['find_regex'] = None 
    if has(options,'path'):
        if options['path'][0] == '!':
            filter['neg_path'] = True
            if type(options['find']) == str:
                filter['path_regex'] = re.compile(r'%s' % 
                    options['path'][1:].decode(encoding), re.IGNORECASE)
            else:
                filter['path_regex'] = re.compile(r'%s' % 
                    options['path'][1:], re.IGNORECASE)

        else:
            filter['neg_path'] = False
            if type(options['find']) == str:
                filter['path_regex'] = re.compile(r'%s' % 
                    options['path'].decode(encoding), re.IGNORECASE)
            else:
                filter['path_regex'] = re.compile(r'%s' % 
                    options['path'], re.IGNORECASE)
    else:
        filter['neg_path'] = False
        filter['path_regex'] = None
    if has(options,'tag'):
        filter['tag_regex'] = []
        for tag in options['tag']:
            if tag[0] == '!':
                neg_tag = True
                if type(tag) == str:
                    filter['tag_regex'].append((neg_tag, re.compile(r'%s' %
                        tag[1:].decode(encoding), re.IGNORECASE)))
                else:
                    filter['tag_regex'].append((neg_tag, re.compile(r'%s' %
                        tag[1:], re.IGNORECASE)))
            else:
                neg_tag = False
                if type(tag) == str:
                    filter['tag_regex'].append((neg_tag, re.compile(r'%s' %
                        tag.decode(encoding), re.IGNORECASE)))
                else:
                    filter['tag_regex'].append((neg_tag, re.compile(r'%s' %
                        tag, re.IGNORECASE)))
    else:
        filter['neg_tag'] = False
        filter['tag_regex'] = []
    return(filter)

def filter_note(filters, note):
    if filters['find_regex']:
        # include the note description in the text 
        s = "%s %s" % (note['description'], " ".join(note['text']))
        r = filters['find_regex'].search(s)
        s_res = (r and r.group(0).strip())
        if (filters['neg_find'] and s_res):
            return(False)
        if not filters['neg_find'] and not s_res:
            return(False)
    if filters['path_regex']:
        r = filters['path_regex'].search(" ".join(note['path']))
        c_res = (r and r.group(0).strip())
        if (filters['neg_path'] and c_res):
            return(False)
        if not filters['neg_path'] and not c_res:
            return(False)
    if filters['tag_regex']: 
        # this will be a list, each element must match 
        t_res = False

        if has(note, 'tags'):
            for neg_tag, tag_regex in filters['tag_regex']:
                for tag in note['tags']:
                    r = tag_regex.search(tag)
                    t_res = (r and r.group(0).strip())
                    if t_res:
                        break
                if (neg_tag and t_res):
                    return(False)
                if not neg_tag and not t_res:
                    return(False)
    return(True)

def datestamps(lines):
    dt_naive = datetime.datetime.now()
    dt_local = dt_naive.replace(tzinfo = tzlocal())
    dt_fmt = dt_local.strftime("%Y-%m-%d %H:%M %Z")
    d_fmt = dt_local.strftime("%Y-%m-%d")
    new_lines = [regex_datetime.sub(dt_fmt, l) for l in lines]
    return([regex_date.sub(d_fmt, l) for l in new_lines])


def noteAdd(rel_path, new_lines):
    new_lines = datestamps(new_lines)
    file = full_path(rel_path)
    if os.path.exists(file):
        backup(file)
    pathname, ext = os.path.splitext(rel_path)
    if ext == ntsenc:
        new_lines = CryptLines(new_lines)
    fo = codecs.open(file, 'a', encoding, 'replace')
    fo.writelines(["%s\n" % x.rstrip('\n') for x in new_lines])
    fo.close()
    fo = codecs.open(file, 'r', encoding, 'replace')
    num_lines = len(fo.readlines())
    fo.close()
    line_beg = num_lines - len(new_lines) + 1
    return(line_beg)

def noteReplace(id, new_lines):
    new_lines = datestamps(new_lines)
    rel_path, line_beg, line_end = id
    file = full_path(rel_path)
    backup(file)
    pathname, ext = os.path.splitext(rel_path)
    if ext == ntsenc:
        encrypt = True
    else:
        encrypt = False
    fo = codecs.open(file, 'r', encoding, 'replace')
    lines = fo.readlines()
    fo.close()
    # leave any blank lines here to keep line numbers correct
    tolines = ["%s\n" % x.rstrip('\n') for x in lines]
    fromlines = ["%s\n" % x.rstrip('\n') for x in lines]
    if encrypt:
        new_lines = CryptLines(new_lines)
    del tolines[line_beg-1:line_end]
    for i in range(len(new_lines)):
        tolines.insert(line_beg -1 + i, "%s\n" % new_lines[i].rstrip('\n'))
    fo = codecs.open(file, 'w', encoding, 'replace')
    fo.writelines(tolines)
    fo.close()
    return(line_beg)

def noteDelete(id):
    rel_path, line_beg, line_end = id
    file = full_path(rel_path)
    backup(file)
    fo = codecs.open(file, 'r', encoding, 'replace')
    lines = fo.readlines()
    fo.close()
    tolines = ["%s\n" % x.rstrip('\n') for x in lines]
    fromlines = ["%s\n" % x.strip('\n') for x in lines]
    del tolines[line_beg-1:line_end]
    # are there any other notes?
    empty = True
    for line in tolines:
        if line.strip():
            empty = False
            break
    if empty:
        os.remove(file)
        res = 2
    else:
        if line_beg > 1:
            tolines.insert(line_beg-1, '\n')
        fo = codecs.open(file, 'w', encoding, 'replace')
        fo.writelines(tolines)
        fo.close()
        res = 1
    return(res)

def backup(file):
    pathname, ext = os.path.splitext(file)
    directory, name = os.path.split(pathname)
    bak = os.path.join(directory, ".%s.bk1" % name)
    if os.path.exists(bak):
        m_secs = os.path.getmtime(bak)
        m_date = datetime.date.fromtimestamp(m_secs)
    else:
        m_date = None
    # only backup if there is no backup or it is at least one day old
    if not m_date or m_date < datetime.date.today():
        for i in range(1, numbaks):
            baknum = numbaks - i
            nextnum = baknum + 1
            bakname = os.path.join(directory,".%s.bk%d" % (name, baknum))
            if os.path.exists(bakname):
                nextname = os.path.join(directory,".%s.bk%d" % (name,
                 nextnum))
                shutil.move(bakname, nextname)
        if ext == ntsenc:
            # already encoded
            shutil.copy2(file, bak)
        else:
            # plain text, convert and leave original
            ConvertFile(file, bak, True, False)
        return(True)
    return(False)

def edit_note(id):
    if not EDITOR:
        print "The EDITOR variable must be set in %s to use this option." % ntsdir
        return()
    rel_path, line_beg, line_end = id
    orig_file = full_path(rel_path)
    pathname, ext = os.path.splitext(rel_path)
    if ext == ntsenc:
        decrypting = True
    else:
        decrypting = False
    fo = codecs.open(orig_file, 'r', encoding, 'replace')
    f = fo.readlines()
    fo.close()
    lines = []
    for line in f[line_beg-1:line_end]:
        line = line.rstrip('\n')
        if decrypting:
            lines.append(decrypt(line))
        else:
            lines.append(line)
    new_lines = ext_lines(lines)
    if len(new_lines) < 2:
        print "Cancelled. At least 2 lines must be entered."
        return()
    if new_lines[0][0] != '+':
        print "Cancelled. The first line must begin with '+'."
        return()
    noteReplace(id, new_lines)

def edit_file(id):
    if not EDITOR:
        print "The EDITOR variable must be set in %s to use this option." % ntsdir
        return()
    temp = None
    rel_path, line_beg, line_end = id
    file = full_path(rel_path)
    if file:
        backup(file)
        pathname, ext = os.path.splitext(file)
        if ext == ntsenc:  # encoded, need a plain text copy
            temp = "%s%s" % (pathname, ntstxt)
            ConvertFile(file, temp, False, True) # removes orig
            command = EDITCMD % {'e': EDITOR, 'n': line_beg, 'f': temp}
            os.system(command)
            ConvertFile(temp, file, True, True) # removes temp
        else: # plain text
            command = EDITCMD % {'e': EDITOR, 'n': line_beg, 'f': file}
            os.system(command)

def get_note(quicknote = False):
    if EDITOR:
        lines = ext_lines()
    else:
        lines = raw_lines()
    if len(lines) < 2:
        print "Cancelled. At least 2 lines must be entered."
        return()
    if lines[0][0] != '+':
        print "Cancelled. The first line must begin with '+'."
        return()
    if quicknote:
        rp = quick
    else:
        rp = get_relpath()
    if rp:
        noteAdd(rp, lines)

def ext_lines(orig_lines=[]):
    temp = os.path.join(ntsdir, '.temp')
    fo = codecs.open(temp, 'w', encoding, 'replace')
    if orig_lines:
        for line in orig_lines:
            fo.write("%s\n" % line)
    else:
        fo.write("+ ")
    fo.close()
    command = EDITCMD % {'e': EDITOR, 'n': '1', 'f': temp}
    os.system(command)
    fo = codecs.open(temp, 'r', encoding, 'replace')
    lines = fo.readlines()
    fo.close()
    os.remove(temp)
    return(lines)

def raw_lines():
    lst = []
    n = 0
    again = True
    print """\
----------------------------------------------------------------------
  Enter a note consisting of two or more lines. Begin the first, title 
  line with '+ '. Subsequent lines form the body of the note. 

  Occurrences of !d found anywhere in the note will be replaced with 
  the current date and time. 

  Enter a line containing only '!' to end the note. 
----------------------------------------------------------------------"""
    while again:
        l = raw_input('> ')
        l = l.decode(encoding)
        m = regex_end.match(l)
        if m:
            again = False
        else:
            lst.append(l)
    return(lst)


def get_opts():
    from optparse import OptionParser, OptionGroup
    usage = """usage: %prog [options] [args]"""
    parser = OptionParser(version = "nts %s" % version,
        description = """Manage notes using simple text files. (C) %s Daniel A Graham.""" % copyright, usage=usage)
    args = OptionGroup(parser, "Args", "numbers and ranges of note numbers to display, e.g., '10 14:16' would limit the display to notes numbered 10, 14, 15 and 16.")
    parser.add_option_group(args)
    parser.add_option("-o", action = "store",
        dest='outlineby',  choices = ['p', 't'], default = 'p',
        help = """An element from [p, t] where:	                          \n
        p: outline by path						  \n
        t: outline by tag						  \n
        Default: %default.""")
    parser.add_option("-p", action = "store",
        dest='path',
        help = """Regular expression. Include items with paths matching PATH (ignoring case). Prepend an exclamation mark, i.e., use !PATH rather than PATH, to include items which do NOT have paths matching PATH.""")
    parser.add_option("-t", action = "append",
        dest='tag',
        help = """Regular expression. Include items with tags matching TAG (ignoring case). Prepend an exclamation mark, i.e., use !TAG rather than TAG, to include items which do NOT have tags matching TAG.""")
    parser.add_option("-f", 
        action="store",
        dest='find',
        help = """Regular expression. Include items containing FIND (ignoring case) in the note text. Prepend an exclamation mark, i.e., use !FIND rather than FIND, to include notes which do NOT have note texts matching FIND.""")
    parser.add_option("-d",
        action="store", dest="display", 
        choices = map(str,range(1,8)), default = '1', 
        help="""An integer from [1, ..., 7] which is the sum of one or   \n 
        more of the following:                                           \n
        1: note title                                               \n 
        2: note body                                                \n
        4: note id and tags                                         \n
        Default: %default.""")
    parser.add_option("-l",
        action="store", dest="lines", type=int, default=0,
        help="""If LINES is 0 or if the note body contains no more than LINES + 1 lines, show the entire body of the note. Else show the first LINES lines and append a line showing the number of omitted lines. Default: %default.""")
    parser.add_option("-n",
        action="store", dest="number", 
        choices = ['0','1'], default = '1', 
        help="""0: hide item numbers; 1: Show item numbers.           \n
        Default: %default.""")
    parser.add_option("-e",
        action="store", dest="edit", 
        type=int, 
        help="""If there is a note numbered EDIT among those which satisfy the current filters, then open that note for editing using EDITOR.""")
    parser.add_option("-E",
        action="store", dest="editfile", 
        type=int, 
        help="""If there is a note numbered EDITFILE among those which satisfy the current filters, then open the file containing that note at the beginning line of the note for editing using EDITOR.""")
    parser.add_option("-r",
        action="store", dest="remove", 
        type=int, 
        help="""If there is a note numbered REMOVE among those which satisfy the current filters, then remove that note after prompting for confirmation.""")
    parser.add_option("--tag_usage", action = "store_true",
        dest='tag_usage',
        help = """Print a report showing the number of uses for each tag and exit.""")
    parser.add_option("-a", action = "store_true",
        dest='add',
        help = """Add a new note.""")
    parser.add_option("-q", action = "store_true",
        dest='quick',
        help = """Add a quick note.""")
    parser.add_option("-N", action = "store_true",
        dest='newer',
        help = """Check for a newer version of nts and exit.""")

    (opts, args) = parser.parse_args()
    opts.__dict__['args'] = args
    return(opts.__dict__)

def load_data(hofh = {}):
    global id2note, id2pth, rel2pth, rel2ids, tag2ids, warnings
    id2note, id2pth, rel2pth, rel2ids, tag2ids, hofh = process_data(hofh)
    warnings = log_messages(messages)
    return(hofh)

def log_messages(messages):
    f = os.path.join(ntsexport, 'messages.log')
    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    if messages:
        message = "\n".join(messages)
        ret = """\
%s 
%d files did not load properly. Details were written to:
    %s
%s""" % ("="*60, len(messages),f, "="*60)
    else:
        message = "Currently there are no messages."
        ret = ""
    fo = codecs.open(f, 'w', encoding, 'replace')
    fo.write("%s\nlast modified: %s\n%s\n" % (f,now, "="*30))
    f = fo.write(message)
    fo.close()
    return(ret)

def main():
    global options, filters
    options = get_opts()
    filters = get_filters(options)
    if options['newer']:
        ok, res = get_newer()
        print "%s\n%s\n%s" % ("="*70, res, "="*70)
        return()
    if options['quick']:
        get_note(True)
        return()
    if options['add']:
        get_note()
        return()
    load_data()
    if options['tag_usage']:
        print "tag usage report"
        print "%s" % '-'*40
        keys = tag2ids.keys()
        keys.sort()
        for key in keys:
            print "%6d: %s" % (len(tag2ids[key]), key)
        return()
    lofl = make_tree(options, filters)
    if warnings:
        print(warnings)
    print_tree(lofl)

if __name__ == '__main__':
    main()